@@ -12,6 +12,8 @@
 // See the License for the specific language governing permissions and
 // limitations under the License.
 
+use std::collections::HashSet;
+
 use proc_macro2::Span;
 use syn::{
     parse::{Parse, ParseStream},
@@ -53,7 +55,7 @@ impl Parse for UnsafePolicy {
 /// Allowlist configuration.
 #[derive(Hash, Debug)]
 pub enum Allowlist {
-    Unspecified,
+    Unspecified(Vec<String>),
     All,
     Specific(Vec<String>),
 }
@@ -61,8 +63,12 @@ pub enum Allowlist {
 impl Allowlist {
     pub(crate) fn push(&mut self, item: LitStr) -> ParseResult<()> {
         match self {
-            Allowlist::Unspecified => {
-                *self = Allowlist::Specific(vec![item.value()]);
+            Allowlist::Unspecified(ref mut uncommitted_list) => {
+                let new_list = uncommitted_list
+                    .drain(..)
+                    .chain(std::iter::once(item.value()))
+                    .collect();
+                *self = Allowlist::Specific(new_list);
             }
             Allowlist::All => {
                 return Err(syn::Error::new(
@@ -89,10 +95,16 @@ impl Allowlist {
 
 impl Default for Allowlist {
     fn default() -> Self {
-        Allowlist::Unspecified
+        Allowlist::Unspecified(Vec::new())
     }
 }
 
+#[derive(Hash, Debug)]
+pub struct Subclass {
+    pub superclass: String,
+    pub subclass: Ident,
+}
+
 #[derive(Hash, Debug)]
 pub struct IncludeCppConfig {
     pub inclusions: Vec<String>,
@@ -104,6 +116,8 @@ pub struct IncludeCppConfig {
     blocklist: Vec<String>,
     exclude_utilities: bool,
     mod_name: Option<Ident>,
+    pub rust_types: Vec<Ident>,
+    pub subclasses: Vec<Subclass>,
 }
 
 impl Parse for IncludeCppConfig {
@@ -117,14 +131,16 @@ impl Parse for IncludeCppConfig {
         let mut parse_only = false;
         let mut exclude_impls = false;
         let mut unsafe_policy = UnsafePolicy::AllFunctionsUnsafe;
-        let mut allowlist = Allowlist::Unspecified;
+        let mut allowlist = Allowlist::default();
         let mut blocklist = Vec::new();
         let mut pod_requests = Vec::new();
+        let mut rust_types = Vec::new();
         let mut exclude_utilities = false;
         let mut mod_name = None;
+        let mut subclasses = Vec::new();
 
         while !input.is_empty() {
-            let has_hexathorpe = input.parse::<Option<syn::Token![#]>>()?.is_some();
+            let has_hexathorpe = input.parse::<Option<syn::token::Pound>>()?.is_some();
             let ident: syn::Ident = input.parse()?;
             if has_hexathorpe {
                 if ident != "include" {
@@ -133,7 +149,7 @@ impl Parse for IncludeCppConfig {
                 let hdr: syn::LitStr = input.parse()?;
                 inclusions.push(hdr.value());
             } else {
-                input.parse::<Option<syn::Token![!]>>()?;
+                input.parse::<Option<syn::token::Bang>>()?;
                 if ident == "generate" {
                     let args;
                     syn::parenthesized!(args in input);
@@ -155,6 +171,27 @@ impl Parse for IncludeCppConfig {
                     syn::parenthesized!(args in input);
                     let generate: syn::LitStr = args.parse()?;
                     blocklist.push(generate.value());
+                } else if ident == "rust_type" {
+                    let args;
+                    syn::parenthesized!(args in input);
+                    let ident: syn::Ident = args.parse()?;
+                    rust_types.push(ident);
+                } else if ident == "subclass" {
+                    let args;
+                    syn::parenthesized!(args in input);
+                    let superclass: syn::LitStr = args.parse()?;
+                    if superclass.value().contains("::") {
+                        return Err(syn::Error::new(
+                            superclass.span(),
+                            "at this time, superclasses can't be in namespaces - see https://github.com/google/autocxx/issues/599",
+                        ));
+                    }
+                    args.parse::<syn::token::Comma>()?;
+                    let subclass: syn::Ident = args.parse()?;
+                    subclasses.push(Subclass {
+                        superclass: superclass.value(),
+                        subclass,
+                    });
                 } else if ident == "parse_only" {
                     parse_only = true;
                     swallow_parentheses(&input, &ident)?;
@@ -194,10 +231,12 @@ impl Parse for IncludeCppConfig {
             parse_only,
             exclude_impls,
             pod_requests,
+            rust_types,
             allowlist,
             blocklist,
             exclude_utilities,
             mod_name,
+            subclasses,
         })
     }
 }
@@ -252,9 +291,21 @@ impl IncludeCppConfig {
                     .iter()
                     .chain(self.pod_requests.iter())
                     .cloned()
-                    .chain(self.active_utilities()),
+                    .chain(self.active_utilities())
+                    .chain(
+                        self.subclasses
+                            .iter()
+                            .map(|sc| {
+                                [
+                                    format!("{}Cpp", sc.subclass),
+                                    sc.subclass.to_string(),
+                                    sc.superclass.clone(),
+                                ]
+                            })
+                            .flatten(),
+                    ),
             )),
-            Allowlist::Unspecified => unreachable!(),
+            Allowlist::Unspecified(_) => unreachable!(),
         }
     }
 
@@ -301,6 +352,16 @@ impl IncludeCppConfig {
                 .unwrap_or_else(|| "default".into())
         )
     }
+
+    pub fn is_rust_type(&self, id: &Ident) -> bool {
+        self.rust_types.contains(id)
+    }
+
+    pub fn superclasses(&self) -> impl Iterator<Item = &String> {
+        let mut uniquified = HashSet::new();
+        uniquified.extend(self.subclasses.iter().map(|sc| &sc.superclass));
+        uniquified.into_iter()
+    }
 }
 
 #[cfg(test)]
